# Estimation of the bit-security of the signature scheme TESLA
# Notations follow the paper, see https://cryptojedi.org/papers/#tesla
#
# To estimate the bit-security of the parameter set given in the paper, i.e., 
# n=416, m=800, q=2**27-39, sigma=114, d=24, omega=20, B=2**23-1,kappa=256, 
# call 
# 'estimate(n=416,  q=2**27-39, sigma=114, d=24, omega=20, B=2**23-1,kappa=256,m=800)'
# to estimate the bit-security in case an optimal choice of parameters were given 
# call 
# 'estimate(n=416,  q=2**27-39, sigma=114, d=24, omega=20, B=2**23-1,kappa=256)'
#
# Main Functions:
# 		estimate()		 	estimates bit-security
# 		get_standard_embedding_delta	estimates Hermite delta under standard embedding attack
# 		get_dual_embedding_delta()	estimates Hermite delta under dual embedding attack
#		get_decoding_delta()		estimates Hermite delta under decoding attack

#
# Auxilary functions:
#		getBitSecurity()		computes the bit-security from a given run time
#
# 		getRunTimeBKZ2()		computes the run time of BKZ2 by estimations of Linder and Peikert
# 		getRunTimeDecodingAttack() 	computes the run time for the decoding attack
#		getRunTimeBKZandNP()

#
# 		getBalancedTimeBKZDecoding()	returns run time of attack such that the run time of decoding and BKZ is balanced 
#		getGramSchmidftBasis()		returns a list consisting if the lenghts of the gram-schmidt vectors 
# 		getSizeBalancedDecoding()	finds value of size such that the run time of decoding and BKZ is balanced 
#
#		calculateD()
#		calculateD()
#		getEnum()
# 		getAlpha()			returns the constant for the GSA ##
# 		getDet()			returns the determinant of the lattice
# 		get_parallel_NP_time()
#		getSuccesP			returns the success probability of nearest planes for LWE with gaussian error
#		myErf()		
#		
################################################################################
#### Main Function #############################################################
#_______________________________________________________________________________
# Computes the bit-security of the signature schemem TESLA
#
# Input: n, m, q, sigma, d, omega, B,kappa
# Output: list l=[embedding,smallm,decoding], bit-securities of attacks
#
def estimate(n,q, sigma, d, omega, B, kappa, m=-1):

    deltaMap = {}
  
    # Standard Embedding Attack
    # no m given, i.e., use optimal m
    if m < 0:
        standard_embedding_delta = get_standard_embedding_delta(n, q, sigma, tau=0.4)
    # considering the size of m
    else: 
        standard_embedding_delta = get_standard_embedding_delta(n, q, sigma, tau=0.4,m=m)
    ListOfBitSecurity=[getBitSecurity(getRunTimeBKZ2(standard_embedding_delta))]   
 
    # Dual Embedding Attack 
    # no m given, i.e., use optimal m
    if m < 0:
        dual_embedding_delta = get_dual_embedding_delta(n=n, q=q, sigma=sigma)
    # m is given
    else: 
        dual_embedding_delta = get_dual_embedding_delta(n=n, q=q, sigma=sigma, m=m)
    ListOfBitSecurity.append(getBitSecurity(getRunTimeBKZ2(dual_embedding_delta)))

    # Decoding
    # no m given, i.e., use optimal m
    if m < 0:
        decoding_delta = get_decoding_delta(n=n, q=q, s = sigma/sqrt(2*pi))
        size = getSizeBalancedDecoding(n, q, decoding_delta)
        decoding_time = getRunTimeDecodingAttack(n, q, decoding_delta, size, s=sigma/sqrt(2*pi))
    # considering the size of m
    else:
        decoding_delta = get_decoding_delta(n=n, q=q, s = sigma/sqrt(2*pi), max_m = m)
        size = getSizeBalancedDecoding(n, q, decoding_delta)
        decoding_time = getRunTimeDecodingAttack(n, q, decoding_delta, size, s=sigma/sqrt(2*pi), max_m = m)
    ListOfBitSecurity.append(getBitSecurity(decoding_time)) 
            
    return ListOfBitSecurity

##
#### Attacks ###################################################################
##
## Standard Embedding Attack ##
#_______________________________________________________________________________
def get_standard_embedding_delta(n, q, sigma_e, tau=0.4,m=-1):
    if m<0:
        return RR(exp( (log(q) - log(sigma_e/sqrt(2*pi)*tau*sqrt(2*pi*e)))^2 / (4*n*log(q)) ))
    else:
        return RR((q**(1-n/m) / ( sqrt(2*e*pi)*tau*sigma_e/sqrt(2*pi)) )**(1/m) )

##
## Dual Embedding Attack ##
#_______________________________________________________________________________
def get_dual_embedding_delta(n, q, sigma, m=-1, tau=0.4):
    if m < 0:
        max_delta = 1
        optimal_m = 0
        for m in range(n+10, 2*n):
            delta = get_dual_embedding_delta(n=n, q=q, sigma=sigma, m=m)
            if delta > max_delta:
                max_delta = delta
                optimal_m = m
        return max_delta
    dim = n+m+1
    gaussian_heuristic = RR(  gamma(1+dim/2)**(1/dim) / sqrt(pi) * q**(m/dim)  )
    norm = RR(sqrt((n+m)*sigma**2 + 1))

    return RR(( gaussian_heuristic/(norm*tau) )**(1/dim))

##
## Decoding Attack ##
#_______________________________________________________________________________
def get_decoding_delta(n, q, s=0, errorSet=[], lx = 1.001, rx=1.02, ddelta = 0.00001, maxCount = 30, max_m = -1):
    mx = (lx + rx)/2

    flx = getBitSecurity(getBalancedTimeBKZDecoding(n=n, q=q, s=s, delta=lx, max_m=max_m))
    frx = getBitSecurity(getBalancedTimeBKZDecoding(n=n, q=q, s=s, delta=rx, max_m=max_m))
    fmx = getBitSecurity(getBalancedTimeBKZDecoding(n=n, q=q, s=s, delta=mx, max_m=max_m))

    try:
      deltaMap
    except NameError:
      deltaMap = {}

    deltaMap[lx] = (flx)
    deltaMap[mx] = (fmx)
    deltaMap[rx] = (frx)

    count = 0
    while (rx-lx > ddelta and count < maxCount):
        if (rx  - mx > mx - lx):
            nx = 1/2 * (mx + rx)
            balanced_time = getBalancedTimeBKZDecoding(n=n, q=q, s=s, delta=nx, max_m=max_m)
            fnx = getBitSecurity(balanced_time)

            deltaMap[nx] = (fnx)
            if (fnx < fmx):
                lx = mx
                flx = fmx
                mx = nx
                fmx = fnx
            else:
                rx = nx
                frx = fnx
        else:
            nx = 1/2 * (lx + mx)
            balanced_time = getBalancedTimeBKZDecoding(n=n, q=q, s=s, delta=nx, max_m=max_m)
            fnx = getBitSecurity(balanced_time)
            if (fmx < fnx):
                lx = nx
                flx = fnx
            else:
                rx = mx
                frx = fmx
                mx = nx
                fmx = fnx
        count = count+1
    if (count == maxCount):
        print "Abort because the maximal numbers iterations was completed."
#        print lx,mx,rx
#        print flx, fmx, frx
#        print intN, intM, intQ, intS, intDelta
#        print
    return mx
		
################################################################################
#### Auxiliary Functiony #######################################################
#_______________________________________________________________________________
# computes the bit-security from a given run time
def getBitSecurity(runtime):
    return RR(log(runtime*2.3e9)/log(2))

#_______________________________________________________________________________
def getRunTimeBKZ2(delta):
    return 2**(1.8/log(delta, base=2) - 110)

#_______________________________________________________________________________
def getRunTimeDecodingAttack(n, q, delta, size, s, m=-1, max_m = -1):
    if m == -1:
        m = round(sqrt(n * log(q) / log(delta)))
        if max_m > 0:
            m = min(m, max_m)
    b = getGramSchmidtBasis(delta, q, n, m)
    d = calculateD(size,b)
    return RR(getRunTimeBKZandNP(n=n, m=m, q=q, s=s, delta=delta, b=b, d=d))

#_______________________________________________________________________________
def getRunTimeBKZandNP(n, m, q, s, delta, b, d):
    BKZ = getRunTimeBKZ2(delta)
    NP = get_parallel_NP_time(m)
    enum = prod(i for i in d)
    p = getSuccesP(m,d,b,s)
    return (BKZ + NP*enum)/p

#_______________________________________________________________________________
def getBalancedTimeBKZDecoding(n, q, delta, s=0, max_m = -1):
    m = round(sqrt(n * log(q) / log(delta))) # optimal number of samples
    if max_m > 0:
        m = min(m, max_m)
    b = getGramSchmidtBasis(delta, q, n, m)
    size = getSizeBalancedDecoding(n, q, delta)
    d = calculateD(size, b)
    time = getRunTimeBKZandNP(n=n, m=m, q=q, s=s, delta=delta, b=b, d=d)
    return time

#_______________________________________________________________________________
def getGramSchmidtBasis(delta,q,n,m):
    alpha = getAlpha(delta,m)
    det = getDet(n,m,q)
    b_i = RR(delta**m*det^(1/m))
    b = []
    for i in range(m):
        b.append(b_i)
        b_i = b_i * alpha
    return b

#_______________________________________________________________________________
def getSizeBalancedDecoding(n, q, delta):
    m = round(sqrt(n * log(q) / log(delta)))
    b = getGramSchmidtBasis(delta, q, n, m)
    enum = getRunTimeBKZ2(delta) / get_parallel_NP_time(m)
    lx = b[len(b)-1]
    rx = b[0]
    flx = getEnum(lx, b) - enum
    frx = getEnum(rx, b) - enum
    for i in range(1000):
        mx = ((lx + rx) / 2)
        fmx = getEnum(mx, b) - enum
        if fmx == 0 or rx / lx < 1+2**(-20): 
            if RR(enum / getEnum(mx, b)) > 2:
                print "war nicht in der lage enum gut zu approximieren", RR(enum / getEnum(mx, b))
            return mx
        if fmx < 0:
            lx = mx
            flx = fmx
        else:
            rx = mx
            frx = fmx
    print "konnte in der angegebenen Anzahl an runden nicht die richtige size finden"

#_______________________________________________________________________________
def calculateD(size, b):
    if round(size/b[len(b)-1]) < 1:
        return [1 for i in range(len(b))]
    d = [max(1, round(size / x)) for x in b]
    return d
#_______________________________________________________________________________
def getEnum(size, b):
    return prod(i for i in calculateD(size, b))

#_______________________________________________________________________________
def getAlpha(delta, m):
    return delta^(-2*m/(m-1))

#_______________________________________________________________________________
def getDet(n,m,q):
    return q^(m-n)

#_______________________________________________________________________________  
def get_parallel_NP_time(m):
    cores = 2**21
    return 2**(-9) / cores

#_______________________________________________________________________________
def getSuccesP(m,d,b,s):
    return RR(prod(myErf(d[i]*b[i]*sqrt(pi)/2/s) for i in range(m)))

#_______________________________________________________________________________
def myErf(x):
    if x > 100:
        return 1
    return erf(x)

